unit SysUtil2;
{$define widechar}
{$ifdef widechar}{$h+}{$else}{$H-}{$endif}
interface
uses Windows;
const
{ File open modes }
  fmOpenRead      =$0000;
  fmOpenWrite     =$0001;
  fmOpenReadWrite =$0002;
  fmShareCompat   =$0000;
  fmShareExclusive=$0010;
  fmShareDenyWrite=$0020;
  fmShareDenyRead =$0030;
  fmShareDenyNone =$0040;

{ File attribute constants }
  faReadOnly =$00000001;
  faHidden   =$00000002;
  faSysFile  =$00000004;
  faVolumeID =$00000008;
  faDirectory=$00000010;
  faArchive  =$00000020;
  faAnyFile  =$0000003F;

  ReadOnly =$00000001;
  Hidden   =$00000002;
  SysFile  =$00000004;
  VolumeID =$00000008;
  Directory=$00000010;
  Archive  =$00000020;
  AnyFile  =$0000003F;

{ File mode magic numbers }
  fmClosed=$D7B0;
  fmInput =$D7B1;
  fmOutput=$D7B2;
  fmInOut =$D7B3;

type
  WordRec=packed record
   Lo,Hi:Byte;
  end;
  LongRec=packed record
   Lo,Hi:Word;
  end;
  TMethod=record
   Code,Data:Pointer;
  end;
  PByteArray=^TByteArray;
  TByteArray=array[0..32767] of Byte;
  PWordArray=^TWordArray;
  TWordArray=array[0..16383] of Word;
  TProcedure=procedure;
  TFileName=string;
  TSearchRec=record
    Time:Integer;
    Size:Integer;
    Attr:Integer;
    Name:TFileName;
    ExcludeAttr:Integer;
    FindHandle:THandle;
    FindData:TWin32FindData;
  end;
  searchrec=tsearchrec;
  TFileRec=record
    Handle:Integer;
    Mode:Integer;
    RecSize:Cardinal;
    Private:array[1..28] of Byte;
    UserData:array[1..32] of Byte;
    Name:array[0..259] of Char;
  end;
  PTextBuf=^TTextBuf;
  TTextBuf=array[0..127] of Char;
  TTextRec=record
    Handle:Integer;
    Mode:Integer;
    BufSize:Cardinal;
    BufPos:Cardinal;
    BufEnd:Cardinal;
    BufPtr:PChar;
    OpenFunc:Pointer;
    InOutFunc:Pointer;
    FlushFunc:Pointer;
    CloseFunc:Pointer;
    UserData:array[1..32] of Byte;
    Name:array[0..259] of Char;
    Buffer:TTextBuf;
  end;

const

{ Empty string and null string pointer. These constants are provided for
  backwards compatibility only. }
  EmptyStr:string='';
  NullStr:PString=@EmptyStr;

{$h-}
type PathStr=String[255];
Function  GetEnv(envvar:string):string;
Function  FExpand(const path:pathstr):pathstr;

{$ifdef widechar}{$h+}{$endif}

function AllocMem(Size:Cardinal):Pointer;
procedure AddExitProc(Proc:TProcedure);
function NewStr(const S:string):PString;
procedure DisposeStr(P:PString);
procedure AssignStr(var P:PString;const S:string);
procedure AppendStr(var Dest:string;const S:string);
{$ifdef widechar}
function UpperCase(const S:string):string;
function LowerCase(const S:string):string;
{$endif}
function CompareStr(const S1,S2:string):Integer;
function CompareText(const S1,S2:string):Integer;
{$ifdef widechar}
function AnsiUpperCase(const S:string):string;
function AnsiLowerCase(const S:string):string;
function AnsiCompareStr(const S1,S2:string):Integer;
function AnsiCompareText(const S1,S2:string):Integer;
{$endif}
function Trim(const S:string):string;
function TrimLeft(const S:string):string;
function TrimRight(const S:string):string;
function QuotedStr(const S:string):string;
{$ifdef widechar}
function AdjustLineBreaks(const S:string):string;
{$endif}
function IsValidIdent(const Ident:string):Boolean;
function StrToIntDef(const S:string;Default:Integer):Integer;

{ LoadStr loads the string resource given by Ident from the application's
  executable file. If the string resource does not exist, an empty string
  is returned. }
function LoadStr(Ident:Integer):string;

{ File management routines }

{ FileOpen opens the specified file using the specified access mode. The
  access mode value is constructed by OR-ing one of the fmOpenXXXX constants
  with one of the fmShareXXXX constants. If the return value is positive,
  the function was successful and the value is the file handle of the opened
  file. A return value of -1 indicates that an error occurred. }
function FileOpen(const FileName:string;Mode:Integer):Integer;

{ FileCreate creates a new file by the specified name. If the return value
  is positive, the function was successful and the value is the file handle
  of the new file. A return value of -1 indicates that an error occurred. }
function FileCreate(const FileName:string):Integer;

{ FileRead reads Count bytes from the file given by Handle into the buffer
  specified by Buffer. The return value is the number of bytes actually
  read;it is less than Count if the end of the file was reached. The return
  value is -1 if an error occurred. }

function FileRead(Handle:Integer;var Buffer;Count:Integer):Integer;

{ FileWrite writes Count bytes to the file given by Handle from the buffer
  specified by Buffer. The return value is the number of bytes actually
  written, or -1 if an error occurred. }

function FileWrite(Handle:Integer;const Buffer;Count:Integer):Integer;

{ FileSeek changes the current position of the file given by Handle to be
  Offset bytes relative to the point given by Origin. Origin=0 means that
  Offset is relative to the beginning of the file, Origin=1 means that
  Offset is relative to the current position, and Origin=2 means that
  Offset is relative to the end of the file. The return value is the new
  current position, relative to the beginning of the file, or -1 if an error
  occurred. }

function FileSeek(Handle,Offset,Origin:Integer):Integer;

{ FileClose closes the specified file. }

procedure FileClose(Handle:Integer);

{ FileAge returns the date-and-time stamp of the specified file. The return
  value can be converted to a TDateTime value using the FileDateToDateTime
  function. The return value is -1 if the file does not exist. }

function FileAge(const FileName:string):Integer;

{ FileExists returns a boolean value that indicates whether the specified
  file exists. }

function FileExists(const FileName:string):Boolean;

function FindFirst(const Path:string;Attr:Integer;var F:TSearchRec):Integer;
function FindNext(var F:TSearchRec):Integer;
procedure FindClose(var F:TSearchRec);

{ FileGetDate returns the DOS date-and-time stamp of the file given by
  Handle. The return value is -1 if the handle is invalid. The
  FileDateToDateTime function can be used to convert the returned value to
  a TDateTime value. }

function FileGetDate(Handle:Integer):Integer;

{ FileSetDate sets the DOS date-and-time stamp of the file given by Handle
  to the value given by Age. The DateTimeToFileDate function can be used to
  convert a TDateTime value to a DOS date-and-time stamp. The return value
  is zero if the function was successful. Otherwise the return value is a
  Windows error code. }

function FileSetDate(Handle:Integer;Age:Integer):Integer;

{ FileGetAttr returns the file attributes of the file given by FileName. The
  attributes can be examined by AND-ing with the faXXXX constants defined
  above. A return value of -1 indicates that an error occurred. }

function FileGetAttr(const FileName:string):Integer;

{ FileSetAttr sets the file attributes of the file given by FileName to the
  value given by Attr. The attribute value is formed by OR-ing the
  appropriate faXXXX constants. The return value is zero if the function was
  successful. Otherwise the return value is a Windows error code. }

function FileSetAttr(const FileName:string;Attr:Integer):Integer;

{ DeleteFile deletes the file given by FileName. The return value is True if
  the file was successfully deleted, or False if an error occurred. }

function DeleteFile(const FileName:string):Boolean;

{ RenameFile renames the file given by OldName to the name given by NewName.
  The return value is True if the file was successfully renamed, or False if
  an error occurred. }

function RenameFile(const OldName,NewName:string):Boolean;

{ ChangeFileExt changes the extension of a filename. FileName specifies a
  filename with or without an extension, and Extension specifies the new
  extension for the filename. The new extension can be a an empty string or
  a period followed by up to three characters. }

function ChangeFileExt(const FileName,Extension:string):string;

{ ExtractFilePath extracts the drive and directory parts of the given
  filename. The resulting string is the rightmost characters of FileName,
  up to and including the colon or backslash that separates the path
  information from the name and extension. The resulting string is empty
  if FileName contains no drive and directory parts. }

function ExtractFilePath(const FileName:string):string;

{ ExtractFileDir extracts the drive and directory parts of the given
  filename. The resulting string is a directory name suitable for passing
  to SetCurrentDir, CreateDir, etc. The resulting string is empty if
  FileName contains no drive and directory parts. }

function ExtractFileDir(const FileName:string):string;

{ ExtractFileDrive extracts the drive part of the given filename.  For
  filenames with drive letters, the resulting string is '<drive>:'.
  For filenames with a UNC path, the resulting string is in the form
  '\\<servername>\<sharename>'.  If the given path contains neither
  style of filename, the result is an empty string. }

function ExtractFileDrive(const FileName:string):string;

{ ExtractFileName extracts the name and extension parts of the given
  filename. The resulting string is the leftmost characters of FileName,
  starting with the first character after the colon or backslash that
  separates the path information from the name and extension. The resulting
  string is equal to FileName if FileName contains no drive and directory
  parts. }

function ExtractFileName(const FileName:string):string;

{ ExtractFileExt extracts the extension part of the given filename. The
  resulting string includes the period character that separates the name
  and extension parts. The resulting string is empty if the given filename
  has no extension. }

function ExtractFileExt(const FileName:string):string;

{ ExpandFileName expands the given filename to a fully qualified filename.
  The resulting string consists of a drive letter, a colon, a root relative
  directory path, and a filename. Embedded '.' and '..' directory references
  are removed. }

function ExpandFileName(const FileName:string):string;

{ FileSearch searches for the file given by Name in the list of directories
  given by DirList. The directory paths in DirList must be separated by
  semicolons. The search always starts with the current directory of the
  current drive. The returned value is a concatenation of one of the
  directory paths and the filename, or an empty string if the file could not
  be located. }

function FileSearch(const Name,DirList:string):string;

function DiskFree(Drive:Byte):Integer;
function DiskSize(Drive:Byte):Integer;

{ GetCurrentDir returns the current directory. }

function GetCurrentDir:string;

{ SetCurrentDir sets the current directory. The return value is True if
  the current directory was successfully changed, or False if an error
  occurred. }

function SetCurrentDir(const Dir:string):Boolean;

{ CreateDir creates a new directory. The return value is True if a new
  directory was successfully created, or False if an error occurred. }

function CreateDir(const Dir:string):Boolean;

{ RemoveDir deletes an existing empty directory. The return value is
  True if the directory was successfully deleted, or False if an error
  occurred. }

function RemoveDir(const Dir:string):Boolean;

{ PChar routines }
function StrLen(Str:PChar):Cardinal;
function StrEnd(Str:PChar):PChar;
function StrMove(Dest,Source:PChar;Count:Cardinal):PChar;
function StrCopy(Dest,Source:PChar):PChar;
function StrECopy(Dest,Source:PChar):PChar;
function StrLCopy(Dest,Source:PChar;MaxLen:Cardinal):PChar;
function StrPCopy(Dest:PChar;const Source:string):PChar;
function StrPLCopy(Dest:PChar;const Source:string;
  MaxLen:Cardinal):PChar;
function StrCat(Dest,Source:PChar):PChar;
function StrLCat(Dest,Source:PChar;MaxLen:Cardinal):PChar;
function StrComp(Str1,Str2:PChar):Integer;
function StrIComp(Str1,Str2:PChar):Integer;
function StrLComp(Str1,Str2:PChar;MaxLen:Cardinal):Integer;
function StrLIComp(Str1,Str2:PChar;MaxLen:Cardinal):Integer;
function StrScan(Str:PChar;Chr:Char):PChar;
function StrRScan(Str:PChar;Chr:Char):PChar;
function StrPos(Str1,Str2:PChar):PChar;
function StrUpper(Str:PChar):PChar;
function StrLower(Str:PChar):PChar;
function StrPas(Str:PChar):string;
function StrAlloc(Size:Cardinal):PChar;
function StrBufSize(Str:PChar):Cardinal;
function StrNew(Str:PChar):PChar;
procedure StrDispose(Str:PChar);
procedure Beep;

implementation

procedure DivMod(Dividend:Integer;Divisor:Word;
  var Result,Remainder:Word);
asm
        PUSH    EBX
        MOV     EBX,EDX
        MOV     EDX,EAX
        SHR     EDX,16
        DIV     BX
        MOV     EBX,Remainder
        MOV     [ECX],AX
        MOV     [EBX],DX
        POP     EBX
end;

function AllocMem(Size:Cardinal):Pointer;
begin
  GetMem(Result,Size);
  FillChar(Result^,Size,0);
end;

type
  PExitProcInfo=^TExitProcInfo;
  TExitProcInfo=record
    Next:PExitProcInfo;
    SaveExit:Pointer;
    Proc:TProcedure;
  end;

const
  ExitProcList:PExitProcInfo=nil;

procedure DoExitProc;far;
var
  P:PExitProcInfo;
  Proc:TProcedure;
begin
  P:=ExitProcList;
  ExitProcList:=P^.Next;
  ExitProc:=P^.SaveExit;
  Proc:=P^.Proc;
  Dispose(P);
  Proc;
end;

procedure AddExitProc(Proc:TProcedure);
var
  P:PExitProcInfo;
begin
  New(P);
  P^.Next:=ExitProcList;
  P^.SaveExit:=ExitProc;
  P^.Proc:=Proc;
  ExitProcList:=P;
  ExitProc:=@DoExitProc;
end;

function NewStr(const S:string):PString;
begin
  if S='' then Result:=NullStr else
  begin
    New(Result);
    Result^:=S;
  end;
end;

procedure DisposeStr(P:PString);
begin
  if (P<>nil) and (P^<>'') then Dispose(P);
end;

procedure AssignStr(var P:PString;const S:string);
var
  Temp:PString;
begin
  Temp:=P;
  P:=NewStr(S);
  DisposeStr(Temp);
end;

procedure AppendStr(var Dest:string;const S:string);
begin
  Dest:=Dest+S;
end;

{$ifdef widechar}

function UpperCase(const S:string):string;
var
  Ch:Char;
  L:Integer;
  Source,Dest:PChar;
begin
  L:=Length(S);
  SetLength(Result,L);
  Source:=Pointer(S);
  Dest:=Pointer(Result);
  while L<>0 do
  begin
    Ch:=Source^;
    if (Ch>='a') and (Ch<='z') then Dec(Ch,32);
    Dest^:=Ch;
    Inc(Source);
    Inc(Dest);
    Dec(L);
  end;
end;

function LowerCase(const S:string):string;
var
  Ch:Char;
  L:Integer;
  Source,Dest:PChar;
begin
  L:=Length(S);
  SetLength(Result,L);
  Source:=Pointer(S);
  Dest:=Pointer(Result);
  while L<>0 do
  begin
    Ch:=Source^;
    if (Ch>='A') and (Ch<='Z') then Inc(Ch,32);
    Dest^:=Ch;
    Inc(Source);
    Inc(Dest);
    Dec(L);
  end;
end;

{$endif}

function CompareStr(const S1,S2:string):Integer;assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     ESI,EAX
        MOV     EDI,EDX
        OR      EAX,EAX
        JE      @@1
        MOV     EAX,[EAX-4]
@@1:    OR      EDX,EDX
        JE      @@2
        MOV     EDX,[EDX-4]
@@2:    MOV     ECX,EAX
        CMP     ECX,EDX
        JBE     @@3
        MOV     ECX,EDX
@@3:    CMP     ECX,ECX
        REPE    CMPSB
        JE      @@4
        MOVZX   EAX,BYTE PTR [ESI-1]
        MOVZX   EDX,BYTE PTR [EDI-1]
@@4:    SUB     EAX,EDX
        POP     EDI
        POP     ESI
end;

function CompareText(const S1,S2:string):Integer;assembler;
asm
        PUSH    ESI
        PUSH    EDI
        PUSH    EBX
        MOV     ESI,EAX
        MOV     EDI,EDX
        OR      EAX,EAX
        JE      @@0
        MOV     EAX,[EAX-4]
@@0:    OR      EDX,EDX
        JE      @@1
        MOV     EDX,[EDX-4]
@@1:    MOV     ECX,EAX
        CMP     ECX,EDX
        JBE     @@2
        MOV     ECX,EDX
@@2:    CMP     ECX,ECX
@@3:    REPE    CMPSB
        JE      @@6
        MOV     BL,BYTE PTR [ESI-1]
        CMP     BL,'a'
        JB      @@4
        CMP     BL,'z'
        JA      @@4
        SUB     BL,20H
@@4:    MOV     BH,BYTE PTR [EDI-1]
        CMP     BH,'a'
        JB      @@5
        CMP     BH,'z'
        JA      @@5
        SUB     BH,20H
@@5:    CMP     BL,BH
        JE      @@3
        MOVZX   EAX,BL
        MOVZX   EDX,BH
@@6:    SUB     EAX,EDX
        POP     EBX
        POP     EDI
        POP     ESI
end;

{$ifdef widechar}

function AnsiUpperCase(const S:string):string;
var
  Len:Integer;
begin
  Len:=Length(S);
  SetString(Result,PChar(S),Len);
  CharUpperBuff(Pointer(Result),Len);
end;

function AnsiLowerCase(const S:string):string;
var
  Len:Integer;
begin
  Len:=Length(S);
  SetString(Result,PChar(S),Len);
  CharLowerBuff(Pointer(Result),Len);
end;

function AnsiCompareStr(const S1,S2:string):Integer;
begin
  Result:=CompareString(LOCALE_USER_DEFAULT,0,PChar(S1),Length(S1),
    PChar(S2),Length(S2))-2;
end;

function AnsiCompareText(const S1,S2:string):Integer;
begin
  Result:=CompareString(LOCALE_USER_DEFAULT,NORM_IGNORECASE,
    PChar(S1),Length(S1),PChar(S2),Length(S2))-2;
end;

{$endif}

function Trim(const S:string):string;
var
  I,L:Integer;
begin
  L:=Length(S);
  I:=1;
  while (I<=L) and (S[I]<=' ') do Inc(I);
  if I>L then Result:='' else
  begin
    while S[L]<=' ' do Dec(L);
    Result:=Copy(S,I,L-I+1);
  end;
end;

function TrimLeft(const S:string):string;
var
  I,L:Integer;
begin
  L:=Length(S);
  I:=1;
  while (I<=L) and (S[I]<=' ') do Inc(I);
  Result:=Copy(S,I,Maxint);
end;

function TrimRight(const S:string):string;
var
  I:Integer;
begin
  I:=Length(S);
  while (I>0) and (S[I]<=' ') do Dec(I);
  Result:=Copy(S,1,I);
end;

function QuotedStr(const S:string):string;
var
  I:Integer;
begin
  Result:=S;
  for I:=Length(Result) downto 1 do
    if Result[I]='''' then Insert('''',Result,I);
  Result:=''''+Result+'''';
end;

{$ifdef widechar}

function AdjustLineBreaks(const S:string):string;
var
  Source,SourceEnd,Dest:PChar;
  Extra:Integer;
begin
  Source:=Pointer(S);
  SourceEnd:=Source+Length(S);
  Extra:=0;
  while Source<SourceEnd do
  begin
    case Source^ of
      #10:
        Inc(Extra);
      #13:
        if Source[1]=#10 then Inc(Source) else Inc(Extra);
    end;
    Inc(Source);
  end;
  if Extra=0 then Result:=S else
  begin
    Source:=Pointer(S);
    SetString(Result,nil,SourceEnd-Source+Extra);
    Dest:=Pointer(Result);
    while Source<SourceEnd do
      case Source^ of
        #10:
          begin
            Dest^:=#13;
            Inc(Dest);
            Dest^:=#10;
            Inc(Dest);
            Inc(Source);
          end;
        #13:
          begin
            Dest^:=#13;
            Inc(Dest);
            Dest^:=#10;
            Inc(Dest);
            Inc(Source);
            if Source^=#10 then Inc(Source);
          end;
      else
        Dest^:=Source^;
        Inc(Dest);
        Inc(Source);
      end;
  end;
end;

{$endif}

function IsValidIdent(const Ident:string):Boolean;
const
  Alpha=['A'..'Z','a'..'z','_'];
  AlphaNumeric=Alpha+['0'..'9'];
var
  I:Integer;
begin
  Result:=False;
  if (Length(Ident)=0) or not (Ident[1] in Alpha) then Exit;
  for I:=2 to Length(Ident) do if not (Ident[I] in AlphaNumeric) then Exit;
  Result:=True;
end;

function StrToIntDef(const S:string;Default:Integer):Integer;
var
  E:Integer;
begin
  Val(S,Result,E);
  if E<>0 then Result:=Default;
end;

function LoadStr(Ident:Integer):string;
var
  Buffer:array[0..1023] of Char;
begin
  SetString(Result,Buffer,LoadString(HInstance,Ident,Buffer,
    SizeOf(Buffer)));
end;

function FileOpen(const FileName:string;Mode:Integer):Integer;
const
  AccessMode:array[0..2] of Integer=(
    GENERIC_READ,
    GENERIC_WRITE,
    GENERIC_READ or GENERIC_WRITE);
  ShareMode:array[0..4] of Integer=(
    0,
    0,
    FILE_SHARE_READ,
    FILE_SHARE_WRITE,
    FILE_SHARE_READ or FILE_SHARE_WRITE);
begin
  Result:=CreateFile(PChar(FileName),AccessMode[Mode and 3],
    ShareMode[(Mode and $F0) shr 4],nil,OPEN_EXISTING,
    FILE_ATTRIBUTE_NORMAL,0);
end;

function FileCreate(const FileName:string):Integer;
begin
  Result:=CreateFile(PChar(FileName),GENERIC_READ or GENERIC_WRITE,
    0,nil,CREATE_ALWAYS,FILE_ATTRIBUTE_NORMAL,0);
end;

function FileRead(Handle:Integer;var Buffer;Count:Integer):Integer;
begin
  if not ReadFile(Handle,Buffer,Count,Result,nil) then Result:=-1;
end;

function FileWrite(Handle:Integer;const Buffer;Count:Integer):Integer;
begin
  if not WriteFile(Handle,Buffer,Count,Result,nil) then Result:=-1;
end;

function FileSeek(Handle,Offset,Origin:Integer):Integer;
begin
  Result:=SetFilePointer(Handle,Offset,nil,Origin);
end;

procedure FileClose(Handle:Integer);
begin
  CloseHandle(Handle);
end;

function FileAge(const FileName:string):Integer;
var
  Handle:THandle;
  FindData:TWin32FindData;
  LocalFileTime:TFileTime;
begin
  Handle:=FindFirstFile(PChar(FileName),FindData);
  if Handle<>INVALID_HANDLE_VALUE then
  begin
    Windows.FindClose(Handle);
    if (FindData.dwFileAttributes and FILE_ATTRIBUTE_DIRECTORY)=0 then
    begin
      FileTimeToLocalFileTime(FindData.ftLastWriteTime,LocalFileTime);
      if FileTimeToDosDateTime(LocalFileTime,LongRec(Result).Hi,
        LongRec(Result).Lo) then Exit;
    end;
  end;
  Result:=-1;
end;

function FileExists(const FileName:string):Boolean;
begin
  Result:=FileAge(FileName)<>-1;
end;

function FileGetDate(Handle:Integer):Integer;
var
  FileTime,LocalFileTime:TFileTime;
begin
  if GetFileTime(Handle,nil,nil,@FileTime) and
    FileTimeToLocalFileTime(FileTime,LocalFileTime) and
    FileTimeToDosDateTime(LocalFileTime,LongRec(Result).Hi,
      LongRec(Result).Lo) then Exit;
  Result:=-1;
end;

function FileSetDate(Handle:Integer;Age:Integer):Integer;
var
  LocalFileTime,FileTime:TFileTime;
begin
  Result:=0;
  if DosDateTimeToFileTime(LongRec(Age).Hi,LongRec(Age).Lo,LocalFileTime) and
    LocalFileTimeToFileTime(LocalFileTime,FileTime) and
    SetFileTime(Handle,nil,nil,@FileTime) then Exit;
  Result:=GetLastError;
end;

function FileGetAttr(const FileName:string):Integer;
begin
  Result:=GetFileAttributes(PChar(FileName));
end;

function FileSetAttr(const FileName:string;Attr:Integer):Integer;
begin
  Result:=0;
  if not SetFileAttributes(PChar(FileName),Attr) then
    Result:=GetLastError;
end;

function FindMatchingFile(var F:TSearchRec):Integer;
var
  LocalFileTime:TFileTime;
begin
  with F do
  begin
    while FindData.dwFileAttributes and ExcludeAttr<>0 do
      if not FindNextFile(FindHandle,FindData) then
      begin
        Result:=GetLastError;
        Exit;
      end;
    FileTimeToLocalFileTime(FindData.ftLastWriteTime,LocalFileTime);
    FileTimeToDosDateTime(LocalFileTime,LongRec(Time).Hi,
      LongRec(Time).Lo);
    Size:=FindData.nFileSizeLow;
    Attr:=FindData.dwFileAttributes;
    Name:=FindData.cFileName;
  end;
  Result:=0;
end;

function FindFirst(const Path:string;Attr:Integer;
  var F:TSearchRec):Integer;
const
  faSpecial=faHidden or faSysFile or faVolumeID or faDirectory;
begin
  F.ExcludeAttr:=not Attr and faSpecial;
  F.FindHandle:=FindFirstFile(PChar(Path),F.FindData);
  if F.FindHandle<>INVALID_HANDLE_VALUE then
  begin
    Result:=FindMatchingFile(F);
    if Result<>0 then FindClose(F);
  end else
    Result:=GetLastError;
end;

function FindNext(var F:TSearchRec):Integer;
begin
  if FindNextFile(F.FindHandle,F.FindData) then
    Result:=FindMatchingFile(F) else
    Result:=GetLastError;
end;

procedure FindClose(var F:TSearchRec);
begin
  if F.FindHandle<>INVALID_HANDLE_VALUE then
    Windows.FindClose(F.FindHandle);
end;

function DeleteFile(const FileName:string):Boolean;
begin
  Result:=Windows.DeleteFile(PChar(FileName));
end;

function RenameFile(const OldName,NewName:string):Boolean;
begin
  Result:=MoveFile(PChar(OldName),PChar(NewName));
end;

function ChangeFileExt(const FileName,Extension:string):string;
var
  I:Integer;
begin
  I:=Length(FileName);
  while (I>0) and not (FileName[I] in ['.','\',':']) do Dec(I);
  if (I=0) or (FileName[I]<>'.') then I:=MaxInt;
  Result:=Copy(FileName,1,I-1)+Extension;
end;

function ExtractFilePath(const FileName:string):string;
var
  I:Integer;
begin
  I:=Length(FileName);
  while (I>0) and not (FileName[I] in ['\',':']) do Dec(I);
  Result:=Copy(FileName,1,I);
end;

function ExtractFileDir(const FileName:string):string;
var
  I:Integer;
begin
  I:=Length(FileName);
  while (I>0) and not (FileName[I] in ['\',':']) do Dec(I);
  if (I>1) and (FileName[I]='\') and
    not (FileName[I-1] in ['\',':']) then Dec(I);
  Result:=Copy(FileName,1,I);
end;

function ExtractFileDrive(const FileName:string):string;
var
  I,J:Integer;
begin
  if (Length(FileName)>=3) and (FileName[2]=':') then
    Result:=Copy(FileName,1,2)
  else if (Length(FileName)>=2) and (FileName[1]='\') and
    (FileName[2]='\') then
  begin
    J:=0;
    I:=3;
    While (I<Length(FileName)) and (J<2) do
    begin
      if FileName[I]='\' then Inc(J);
      if J<2 then Inc(I);
    end;
    if FileName[I]='\' then Dec(I);
    Result:=Copy(FileName,1,I);
  end else Result:='';
end;

function ExtractFileName(const FileName:string):string;
var
  I:Integer;
begin
  I:=Length(FileName);
  while (I>0) and not (FileName[I] in ['\',':']) do Dec(I);
  Result:=Copy(FileName,I+1,255);
end;

function ExtractFileExt(const FileName:string):string;
var
  I:Integer;
begin
  I:=Length(FileName);
  while (I>0) and not (FileName[I] in ['.','\',':']) do Dec(I);
  if (I>0) and (FileName[I]='.') then
    Result:=Copy(FileName,I,255) else
    Result:='';
end;

function ExpandFileName(const FileName:string):string;
var
  FName:PChar;
  Buffer:array[0..MAX_PATH-1] of Char;
begin
  SetString(Result,Buffer,GetFullPathName(PChar(FileName),SizeOf(Buffer),
    Buffer,FName));
end;

function FileSearch(const Name,DirList:string):string;
var
  I,P,L:Integer;
begin
  Result:=Name;
  P:=1;
  L:=Length(DirList);
  while True do
  begin
    if FileExists(Result) then Exit;
    while (P<=L) and (DirList[P]=';') do Inc(P);
    if P>L then Break;
    I:=P;
    while (P<=L) and (DirList[P]<>';') do Inc(P);
    Result:=Copy(DirList,I,P-I);
    if not (DirList[P-1] in [':','\']) then Result:=Result+'\';
    Result:=Result+Name;
  end;
  Result:='';
end;

function DiskFree(Drive:Byte):Integer;
var
  RootPath:array[0..4] of Char;
  RootPtr:PChar;
  SectorsPerCluster,
  BytesPerSector,
  FreeClusters,
  TotalClusters:Integer;
begin
  RootPtr:=nil;
  if Drive>0 then
  begin
    StrCopy(RootPath,'A:\');
    RootPath[0]:=Char(Drive+$40);
    RootPtr:=RootPath;
  end;
  if GetDiskFreeSpace(RootPtr,SectorsPerCluster,BytesPerSector,
    FreeClusters,TotalClusters) then
    Result:=SectorsPerCluster*BytesPerSector*FreeClusters
  else Result:=-1;
end;

function DiskSize(Drive:Byte):Integer;
var
  RootPath:array[0..4] of Char;
  RootPtr:PChar;
  SectorsPerCluster,
  BytesPerSector,
  FreeClusters,
  TotalClusters:Integer;
begin
  RootPtr:=nil;
  if Drive>0 then
  begin
    StrCopy(RootPath,'A:\');
    RootPath[0]:=Char(Drive+$40);
    RootPtr:=RootPath;
  end;
  if GetDiskFreeSpace(RootPtr,SectorsPerCluster,BytesPerSector,
    FreeClusters,TotalClusters) then
    Result:=SectorsPerCluster*BytesPerSector*TotalClusters
  else Result:=-1;
end;

function GetCurrentDir:string;
var
  Buffer:array[0..MAX_PATH-1] of Char;
begin
  SetString(Result,Buffer,GetCurrentDirectory(SizeOf(Buffer),Buffer));
end;

function SetCurrentDir(const Dir:string):Boolean;
begin
  Result:=SetCurrentDirectory(PChar(Dir));
end;

function CreateDir(const Dir:string):Boolean;
begin
  Result:=CreateDirectory(PChar(Dir),nil);
end;

function RemoveDir(const Dir:string):Boolean;
begin
  Result:=RemoveDirectory(PChar(Dir));
end;

function StrLen(Str:PChar):Cardinal;assembler;
asm
        MOV     EDX,EDI
        MOV     EDI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        MOV     EAX,0FFFFFFFEH
        SUB     EAX,ECX
        MOV     EDI,EDX
end;

function StrEnd(Str:PChar):PChar;assembler;
asm
        MOV     EDX,EDI
        MOV     EDI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        LEA     EAX,[EDI-1]
        MOV     EDI,EDX
end;

function StrMove(Dest,Source:PChar;Count:Cardinal):PChar;assembler;
asm
        PUSH    ESI
        PUSH    EDI
        MOV     ESI,EDX
        MOV     EDI,EAX
        MOV     EDX,ECX
        CMP     EDI,ESI
        JG      @@1
        JE      @@2
        SHR     ECX,2
        REP     MOVSD
        MOV     ECX,EDX
        AND     ECX,3
        REP     MOVSB
        JMP     @@2
@@1:    LEA     ESI,[ESI+ECX-1]
        LEA     EDI,[EDI+ECX-1]
        AND     ECX,3
        STD
        REP     MOVSB
        SUB     ESI,3
        SUB     EDI,3
        MOV     ECX,EDX
        SHR     ECX,2
        REP     MOVSD
        CLD
@@2:    POP     EDI
        POP     ESI
end;

function StrCopy(Dest,Source:PChar):PChar;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     ECX
        MOV     EDI,ESI
        MOV     ESI,EDX
        MOV     EDX,ECX
        MOV     EAX,EDI
        SHR     ECX,2
        REP     MOVSD
        MOV     ECX,EDX
        AND     ECX,3
        REP     MOVSB
        POP     ESI
        POP     EDI
end;

function StrECopy(Dest,Source:PChar):PChar;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     ECX
        MOV     EDI,ESI
        MOV     ESI,EDX
        MOV     EDX,ECX
        SHR     ECX,2
        REP     MOVSD
        MOV     ECX,EDX
        AND     ECX,3
        REP     MOVSB
        LEA     EAX,[EDI-1]
        POP     ESI
        POP     EDI
end;

function StrLCopy(Dest,Source:PChar;MaxLen:Cardinal):PChar;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        MOV     ESI,EAX
        MOV     EDI,EDX
        MOV     EBX,ECX
        XOR     AL,AL
        TEST    ECX,ECX
        JZ      @@1
        REPNE   SCASB
        JNE     @@1
        INC     ECX
@@1:    SUB     EBX,ECX
        MOV     EDI,ESI
        MOV     ESI,EDX
        MOV     EDX,EDI
        MOV     ECX,EBX
        SHR     ECX,2
        REP     MOVSD
        MOV     ECX,EBX
        AND     ECX,3
        REP     MOVSB
        STOSB
        MOV     EAX,EDX
        POP     EBX
        POP     ESI
        POP     EDI
end;

function StrPCopy(Dest:PChar;const Source:string):PChar;
begin
  Result:=StrLCopy(Dest,PChar(Source),255);
end;

function StrPLCopy(Dest:PChar;const Source:string;
  MaxLen:Cardinal):PChar;
begin
  Result:=StrLCopy(Dest,PChar(Source),MaxLen);
end;

function StrCat(Dest,Source:PChar):PChar;
begin
  StrCopy(StrEnd(Dest),Source);
  Result:=Dest;
end;

function StrLCat(Dest,Source:PChar;MaxLen:Cardinal):PChar;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        MOV     EDI,Dest
        MOV     ESI,Source
        MOV     EBX,MaxLen
        CALL    StrEnd
        MOV     ECX,EDI
        ADD     ECX,EBX
        SUB     ECX,EAX
        JBE     @@1
        MOV     EDX,ESI
        CALL    StrLCopy
@@1:    MOV     EAX,EDI
        POP     EBX
        POP     ESI
        POP     EDI
end;

function StrComp(Str1,Str2:PChar):Integer;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     EDI,EDX
        MOV     ESI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     EAX,EAX
        REPNE   SCASB
        NOT     ECX
        MOV     EDI,EDX
        XOR     EDX,EDX
        REPE    CMPSB
        MOV     AL,[ESI-1]
        MOV     DL,[EDI-1]
        SUB     EAX,EDX
        POP     ESI
        POP     EDI
end;

function StrIComp(Str1,Str2:PChar):Integer;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        MOV     EDI,EDX
        MOV     ESI,EAX
        MOV     ECX,0FFFFFFFFH
        XOR     EAX,EAX
        REPNE   SCASB
        NOT     ECX
        MOV     EDI,EDX
        XOR     EDX,EDX
@@1:    REPE    CMPSB
        JE      @@4
        MOV     AL,[ESI-1]
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
@@2:    MOV     DL,[EDI-1]
        CMP     DL,'a'
        JB      @@3
        CMP     DL,'z'
        JA      @@3
        SUB     DL,20H
@@3:    SUB     EAX,EDX
        JE      @@1
@@4:    POP     ESI
        POP     EDI
end;

function StrLComp(Str1,Str2:PChar;MaxLen:Cardinal):Integer;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        MOV     EDI,EDX
        MOV     ESI,EAX
        MOV     EBX,ECX
        XOR     EAX,EAX
        OR      ECX,ECX
        JE      @@1
        REPNE   SCASB
        SUB     EBX,ECX
        MOV     ECX,EBX
        MOV     EDI,EDX
        XOR     EDX,EDX
        REPE    CMPSB
        MOV     AL,[ESI-1]
        MOV     DL,[EDI-1]
        SUB     EAX,EDX
@@1:    POP     EBX
        POP     ESI
        POP     EDI
end;

function StrLIComp(Str1,Str2:PChar;MaxLen:Cardinal):Integer;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        MOV     EDI,EDX
        MOV     ESI,EAX
        MOV     EBX,ECX
        XOR     EAX,EAX
        OR      ECX,ECX
        JE      @@4
        REPNE   SCASB
        SUB     EBX,ECX
        MOV     ECX,EBX
        MOV     EDI,EDX
        XOR     EDX,EDX
@@1:    REPE    CMPSB
        JE      @@4
        MOV     AL,[ESI-1]
        CMP     AL,'a'
        JB      @@2
        CMP     AL,'z'
        JA      @@2
        SUB     AL,20H
@@2:    MOV     DL,[EDI-1]
        CMP     DL,'a'
        JB      @@3
        CMP     DL,'z'
        JA      @@3
        SUB     DL,20H
@@3:    SUB     EAX,EDX
        JE      @@1
@@4:    POP     EBX
        POP     ESI
        POP     EDI
end;

function StrScan(Str:PChar;Chr:Char):PChar;assembler;
asm
        PUSH    EDI
        PUSH    EAX
        MOV     EDI,Str
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     ECX
        POP     EDI
        MOV     AL,Chr
        REPNE   SCASB
        MOV     EAX,0
        JNE     @@1
        MOV     EAX,EDI
        DEC     EAX
@@1:    POP     EDI
end;

function StrRScan(Str:PChar;Chr:Char):PChar;assembler;
asm
        PUSH    EDI
        MOV     EDI,Str
        MOV     ECX,0FFFFFFFFH
        XOR     AL,AL
        REPNE   SCASB
        NOT     ECX
        STD
        DEC     EDI
        MOV     AL,Chr
        REPNE   SCASB
        MOV     EAX,0
        JNE     @@1
        MOV     EAX,EDI
        INC     EAX
@@1:    CLD
        POP     EDI
end;

function StrPos(Str1,Str2:PChar):PChar;assembler;
asm
        PUSH    EDI
        PUSH    ESI
        PUSH    EBX
        MOV     EBX,EAX
        MOV     EDI,EDX
        XOR     AL,AL
        MOV     ECX,0FFFFFFFFH
        REPNE   SCASB
        NOT     ECX
        DEC     ECX
        JE      @@2
        MOV     ESI,ECX
        MOV     EDI,EBX
        MOV     ECX,0FFFFFFFFH
        REPNE   SCASB
        NOT     ECX
        SUB     ECX,ESI
        JBE     @@2
        MOV     EDI,EBX
        LEA     EBX,[ESI-1]
@@1:    MOV     ESI,EDX
        LODSB
        REPNE   SCASB
        JNE     @@2
        MOV     EAX,ECX
        PUSH    EDI
        MOV     ECX,EBX
        REPE    CMPSB
        POP     EDI
        MOV     ECX,EAX
        JNE     @@1
        LEA     EAX,[EDI-1]
        JMP     @@3
@@2:    XOR     EAX,EAX
@@3:    POP     EBX
        POP     ESI
        POP     EDI
end;

function StrUpper(Str:PChar):PChar;assembler;
asm
        PUSH    ESI
        MOV     ESI,Str
        MOV     EDX,Str
@@1:    LODSB
        OR      AL,AL
        JE      @@2
        CMP     AL,'a'
        JB      @@1
        CMP     AL,'z'
        JA      @@1
        SUB     AL,20H
        MOV     [ESI-1],AL
        JMP     @@1
@@2:    XCHG    EAX,EDX
        POP     ESI
end;

function StrLower(Str:PChar):PChar;assembler;
asm
        PUSH    ESI
        MOV     ESI,Str
        MOV     EDX,Str
@@1:    LODSB
        OR      AL,AL
        JE      @@2
        CMP     AL,'A'
        JB      @@1
        CMP     AL,'Z'
        JA      @@1
        ADD     AL,20H
        MOV     [ESI-1],AL
        JMP     @@1
@@2:    XCHG    EAX,EDX
        POP     ESI
end;

function StrPas(Str:PChar):string;
begin
  Result:=Str;
end;

function StrAlloc(Size:Cardinal):PChar;
begin
  Inc(Size,SizeOf(Cardinal));
  GetMem(Result,Size);
  Cardinal(Pointer(Result)^):=Size;
  Inc(Result,SizeOf(Cardinal));
end;

function StrBufSize(Str:PChar):Cardinal;
begin
  Dec(Str,SizeOf(Cardinal));
  Result:=Cardinal(Pointer(Str)^)-SizeOf(Cardinal);
end;

function StrNew(Str:PChar):PChar;
var
  Size:Cardinal;
begin
  if Str=nil then Result:=nil else
  begin
    Size:=StrLen(Str)+1;
    Result:=StrMove(StrAlloc(Size),Str,Size);
  end;
end;

procedure StrDispose(Str:PChar);
begin
  if Str<>nil then
  begin
    Dec(Str,SizeOf(Cardinal));
    FreeMem(Str,Cardinal(Pointer(Str)^));
  end;
end;

procedure ScanBlanks(const S:string;var Pos:Integer);
var
  I:Integer;
begin
  I:=Pos;
  while (I<=Length(S)) and (S[I]=' ') do Inc(I);
  Pos:=I;
end;

function ScanNumber(const S:string;var Pos:Integer;var Number:Word):Boolean;
var
  I:Integer;
  N:Word;
begin
  Result:=False;
  ScanBlanks(S,Pos);
  I:=Pos;
  N:=0;
  while (I<=Length(S)) and (S[I] in ['0'..'9']) and (N<1000) do
  begin
    N:=N*10+(Ord(S[I])-Ord('0'));
    Inc(I);
  end;
  if I>Pos then
  begin
    Pos:=I;
    Number:=N;
    Result:=True;
  end;
end;

{$ifdef widechar}

function ScanString(const S:string;var Pos:Integer;
  const Symbol:string):Boolean;
begin
  Result:=False;
  if Symbol<>'' then
  begin
    ScanBlanks(S,Pos);
    if AnsiCompareText(Symbol,Copy(S,Pos,Length(Symbol)))=0 then
    begin
      Inc(Pos,Length(Symbol));
      Result:=True;
    end;
  end;
end;

{$endif}

function ScanChar(const S:string;var Pos:Integer;Ch:Char):Boolean;
begin
  Result:=False;
  ScanBlanks(S,Pos);
  if (Pos<=Length(S)) and (S[Pos]=Ch) then
  begin
    Inc(Pos);
    Result:=True;
  end;
end;

{ Convert physical address to logical address }

function ConvertAddr(Address:Pointer):Pointer;assembler;
asm
        TEST    EAX,EAX         { Always convert nil to nil }
        JE      @@1
        SUB     EAX,OFFSET TextStart
@@1:
end;

procedure Beep;
begin
  MessageBeep(0);
end;

{$h-}

function upcasea(const s:shortstring):shortstring;
var
 i:longint;
begin
 upcasea[0]:=s[0];
 for i:=1 to length(s) do upcasea[i]:=upcase(s[i]);
end;

Function GetEnv(envvar:string):string;
var
  s:string;
  i:longint;
  hp,p:pchar;
begin
   getenv:='';
   p:=GetEnvironmentStrings;
   hp:=p;
   while hp^<>#0 do
     begin
        s:=strpas(hp);
        i:=pos('=',s);
        if upcasea(copy(s,1,i-1))=upcasea(envvar) then
          begin
             getenv:=copy(s,i+1,length(s)-i);
             break;
          end;
        hp:=hp+strlen(hp)+1;
     end;
   FreeEnvironmentStrings(p);
end;


function fexpand(const path:pathstr):pathstr;
var
   s,pa:pathstr;
   i,j :longint;
begin
   getdir(0,s);
   i:=ioresult;
   pa:=path;
   for i:=1 to length(pa) do
    if pa[i]='/' then pa[i]:='\';
   if (length(pa)>1) and (pa[2]=':') and (pa[1] in ['A'..'Z','a'..'z']) then
     begin
        if (pa[1] in ['a'..'z']) then
         pa[1]:=Chr(Ord(Pa[1])-32);
        getdir(ord(pa[1])-ord('A')+1,s);
        if (ord(pa[0])>2) and (pa[3]<>'\') then
          if pa[1]=s[1] then
            pa:=s+'\'+copy (pa,3,length(pa))
          else
            pa:=pa[1]+':\'+copy (pa,3,length(pa))
     end
   else
     if pa[1]='\' then pa:=s[1]+':'+pa
     else if s[0]=#3 then pa:=s+pa
     else pa:=s+'\'+pa;
 if length(pa)=2 then
  begin
    getdir(byte(pa[1])-64,s);
    i:=ioresult;
    pa:=s;
  end;
   while pos ('\.\',pa)<>0 do
    delete (pa,pos('\.\',pa),2);
   repeat
     i:=pos('\..\',pa);
     if i<>0 then
      begin
        j:=i-1;
        while (j>1) and (pa[j]<>'\') do
         dec (j);
        if pa[j+1]=':' then j:=3;
        delete (pa,j,i-j+3);
      end;
   until i=0;
   i:=pos('\..',pa);
   if i<>0 then
    begin
      if i=length(pa)-2 then
       begin
         j:=i-1;
         while (j>1) and (pa[j]<>'\') do
          dec (j);
         delete (pa,j,i-j+3);
       end;
       pa:=pa+'\';
     end;
   if (length(pa)>0) and (pa[length(pa)]='.') then
    dec(byte(pa[0]));
   if ((length(pa)>3) and (pa[length(pa)]='\')) and (path[length(path)]<>'\') then
    dec(byte(pa[0]));
   if length(path)=2 then pa:=pa+'\';
   fexpand:=pa;
end;

end.